import json

class ChatJoinRequest:
  def __init__(
    self,
    client: "Nexgram.Client",
    from_user: "Nexgram.types.User",
    chat: "Nexgram.types.Chat",
  ):
    from Nexgram.types import User, Chat
    if not isinstance(from_user, User):
      raise Exception("Cry now i won't say what exception is this")
    elif not isinstance(chat, Chat):
      raise Exception("Cry now i won't say what exception is this")
    self._ = "Nexgram.types.ChatJoinRequest" 
    self.id = id
    self.chat = chat
    self.from_user = from_user
    self.client = client
  def __repr__(self):
    from Nexgram import Client
    mf = {"client"}
    def clean(obj):
      if isinstance(obj, Client):
        return None
      if isinstance(obj, dict):
        return {k: clean(v) for k, v in obj.items() if k not in mf}
      if hasattr(obj, "__dict__"):
        return {
          k: clean(v)
          for k, v in obj.__dict__.items()
          if k not in mf and not callable(v)
        }
      return obj
    return json.dumps(clean(self), indent=2, ensure_ascii=False).replace("\\n", "\n")    
    


from Nexgram.types import *

class CreateMessage:
  async def create_message(self, data, update_type='message'):
    frm = data.get('from_user') or data.get('from')
    chat = data.get('chat')
    callback_query_message = data.get('message')
    forward_from = data.get('forward_from')
    forward_from_chat = data.get("forward_from_chat")
    
    if frm:
      from_user = User(
        client=self,
        id=frm.get('id'),
        first_name=frm.get('first_name'),
        last_name=frm.get('last_name'),
        username=frm.get('username'),
        is_bot=frm.get('is_bot'),
      )  
    if chat:
      chat = Chat(
        id=chat.get('id'),
        title=chat.get('title'),
        first_name=chat.get('first_name'),
        last_name=chat.get('last_name'),
        type=chat.get('type'),
        username=chat.get('username'),
      )
    if callback_query_message:
      callback_query_message = await self.create_message(callback_query_message)
    if forward_from:
      forward_from = User(
        client=self,
        id=forward_from.get('id'),
        first_name=forward_from.get('first_name'),
        last_name=forward_from.get('last_name'),
        username=forward_from.get("username"),
        is_bot=forward_from.get("is_bot")
      )
    if forward_from_chat:
      forward_from_chat = Chat(
        id=forward_from_chat.get('id'),
        title=forward_from_chat.get('title'),
        first_name=forward_from_chat.get('first_name'),
        last_name=forward_from_chat.get('last_name'),
        type=forward_from_chat.get('type'),
        username=forward_from_chat.get('username'),
      )
    if update_type == "message":
      return Message(
        client=self,
        id=data.get('message_id') or data.get('id'),
        from_user=from_user,
        chat=chat,
        forward_from=forward_from,
        forward_from_chat=forward_from_chat,
        reply_markup=None,
        caption=data.get('caption'),
        text=data.get('text')
      )
    elif update_type == "callback_query":
      return CallbackQuery(
        client=self,
        id=data.get('id'),
        from_user=from_user,
        message=callback_query_message,
        data=data.get('data')
      )
    elif update_type == "inline_query":
      return InlineQuery(
        client=self,
        id=data.get('id'),
        from_user=from_user,
        query=data.get('query'),
        offset=data.get('offset')
      )


import json

class CallbackQuery:
  def __init__(
    self,
    client: "Nexgram.Client",
    id: int,
    from_user: "Nexgram.types.User",
    message: "Nexgram.types.Message" = None,
    data: str = None,
  ):
    from Nexgram.types import User, Message
    if not isinstance(from_user, User):
      raise Exception("Cry now i won't say what exception is this")
    elif message and not isinstance(message, Message):
      raise Exception("Cry now i won't say what exception is this")
    self._ = "Nexgram.types.CallbackQuery"
    self.id = id
    self.from_user = from_user
    self.message = message if message else None
    self.data = data
    self.client = client
  def __repr__(self):
    from Nexgram import Client
    mf = {"client"}
    def clean(obj):
      if isinstance(obj, Client):
        return None
      if isinstance(obj, dict):
        return {k: clean(v) for k, v in obj.items() if k not in mf}
      if hasattr(obj, "__dict__"):
        return {
          k: clean(v)
          for k, v in obj.__dict__.items()
          if k not in mf and not callable(v)
        }
      return obj
    return json.dumps(clean(self), indent=2, ensure_ascii=False).replace("\\n", "\n")    
  async def answer(self, text: str, show_alert: bool = False, cache_time: int = 0):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    data = {
      "callback_query_id": self.id,
      "text": text,
      "show_alert": show_alert,
      "cache_time": cache_time,
    }
    return await api.post(url+"answerCallbackQuery", data)


from Nexgram.types import *
import asyncio

modes = ['polling', 'webhook', 'none']

class Start:
  async def start(
    self,
    mode: str = 'polling',
    webhook_url: str = None,
    webhook_port: int = None,
  ):
    api, url, log = self.api, self.ApiUrl+"getMe", self.log
    if mode.lower() not in modes:
      raise ValueError(f"Mode must be 'polling' or 'webhook' not '{mode}'")
    self.mode = mode.lower()
    r = await api.get(url)
    if r.get("ok"):
      self.connected = True
      r = r["result"]
      self.me = User(
        client=self,
        id=r['id'],
        first_name=r['first_name'],
        username=r['username'],
        is_self=True,
        is_bot=True,
      )
      log.info(f"Client connected as {self.me.first_name} (@{self.me.username})")
      if mode=='polling' and True:
        asyncio.create_task(self.start_polling())
        log.info("Exp. Feature Started: Loop created.")
      elif mode == "webhook":
        if not webhook_url or not webhook_port:
          raise ValueError("you selected 'webhook' mode. then where is url & port? you should provied it.")
        if webhook_url.endswith('/'): webhook_url = webhook_url[:-1]
        self.webhook_url, self.webhook_port = webhook_url, webhook_port
        loop = asyncio.get_event_loop()
        loop.create_task(self.createWebhook())
      return self.me
    raise ValueError("Failed to connect with your bot token. Please make sure your bot token is correct.")


import json

class User:
  def __init__(
    self,
    client: "Nexgram.Client" = None,
    id: int = 0,
    first_name: str = None,
    last_name: str = None,
    username: str = None,
    is_bot: bool = False,
    is_self: bool = False,
  ):
    from Nexgram import Client
    self._ = "Nexgram.types.User"
    self.id = int(id)
    if client:
      self.is_self = is_self or client.me.id == self.id
      if not isinstance(client, Client):
        raise TypeError("You should only pass 'Nexgram.Client' object in 'client' argument not others!")
      self.client = client
    else: self.is_self = is_self
    self.is_bot = is_bot
    self.first_name = first_name
    if last_name: self.last_name = last_name
    if username: self.username = username
      
  def __repr__(self):
    from Nexgram import Client
    mf = {"client"}
    def clean(obj):
      if isinstance(obj, Client):
        return None
      if isinstance(obj, dict):
        return {k: clean(v) for k, v in obj.items() if k not in mf}
      if hasattr(obj, "__dict__"):
        return {k: clean(v) for k, v in obj.__dict__.items() if k not in mf}
      return obj
    return json.dumps(clean(self), indent=2, ensure_ascii=False).replace("\\n", "\n")
    


import json
from Nexgram.errors import *

class Message:
  def __init__(
    self,
    client: "Nexgram.Client",
    id: int,
    from_user: "Nexgram.types.User",
    chat: "Nexgram.types.Chat",
    reply_to_message: "Nexgram.types.Message" = None,
    forward_from: "Nexgram.types.User" = None,
    forward_from_chat: "Nexgram.types.Chat" = None,
    reply_markup: "Nexgram.types.InlineKeyboardMarkup" = None,
    caption: str = None,
    text: str = None
  ):
    from Nexgram.types import User, Chat
    from Nexgram import Client
    
    if not isinstance(from_user, User): raise InvalidObject("You should pass 'Nexgram.types.User' object in 'from_user' argument not others.")
    if not isinstance(client, Client): raise InvalidObject("You should pass 'Nexgram.Client' object in 'client' argument not others")
    
    self._ = "Nexgram.types.Message"
    self.id = id
    self.from_user = from_user
    if not isinstance(chat, Chat): raise InvalidObject("You should pass 'Nexgram.types.Chat' object in 'chat' argument not others.")
    self.chat = chat
    if reply_to_message:
      if not isinstance(reply_to_message, Message):
        raise InvalidObject("You should pass 'Nexgram.Client.Message' object in 'reply_to_message' argument not others")
      self.reply_to_message = reply_to_message
    if forward_from:
      if not isinstance(forward_from, User): raise TypeError("?.")
      self.forward_from = forward_from
    if forward_from_chat:
      if not isinstance(forward_from_chat, Chat): raise TypeError("?.")
      self.forward_from_chat = forward_from_chat
    if caption: self.caption = caption
    if text: self.text = text
    self.client = client
  
  def __repr__(self):
    from Nexgram import Client
    mf = {"client"}
    def clean(obj):
      if isinstance(obj, Client):
        return None
      if isinstance(obj, dict):
        return {k: clean(v) for k, v in obj.items() if k not in mf}
      if hasattr(obj, "__dict__"):
        return {
          k: clean(v)
          for k, v in obj.__dict__.items()
          if k not in mf and not callable(v)
        }
      return obj
    return json.dumps(clean(self), indent=2, ensure_ascii=False).replace("\\n", "\n")    
  async def reply(self, text: str, reply_markup = None,parse_mode: str = None):
    client = self.client
    await client.send_message(
      chat_id=self.chat.id,
      text=text,
      reply_markup=reply_markup,
      reply_to_message_id=self.id,
      parse_mode=parse_mode,
    )
  async def delete(self):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    return await api.post(url+"deleteMessage", {"chat_id": self.chat.id, "message_id": self.id})
  
  async def forward(self, chat_id):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    return await client.forward_messages(chat_id, self.chat.id, self.id)
  
  async def copy(self, chat_id, caption=None, parse_mode=None):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    return await client.copy_messages(chat_id, self.chat.id, self.id, caption=caption, parse_mode=parse_mode)


import logging
import httpx
import aiohttp
import asyncio
from .methods import *
from .errors import *
from .types import *
from .api import Api

log = logging.getLogger(__name__)
clients = []

class Client(Methods):
  def __init__(
    self,
    name: str,
    bot_token: str,
  ):
    self.name = name
    self.bot_token = bot_token
    self.connected = False
    self.me = None
    self.offset = 0
    self.polling = False
    self.ApiUrl = f"https://api.telegram.org/bot{self.bot_token}/"
    self.api = Api()
    self.log = log
    self.mode = None
    clients.append(self)
    # Decorators --
    self.on_message_listeners = {}
    self.on_disconnect_listeners = {}
    self.on_callback_query_listeners = {}
    self.on_inline_query_listeners = {}
    self.on_listeners = {}


from Nexgram.errors import *
from Nexgram.types import *
import asyncio

class Dispatch:
  async def __dispatch_helper(self, src, update, update_type='message'):
    try:
      m = update.get(update_type)
      message = await self.create_message(m, update_type=update_type)
      for x in src:
        asyncio.create_task(self.call(src, x, self, message))
    except Exception as e:
      self.log.error(f"[DispatchUpdate] Line 13: {e}")
  
  async def dispatch_update(self, update):
    log = self.log
    for gf in self.on_listeners:
      asyncio.create_task(gf(update))
    if update.get('message'):
      update_type, src = "message", self.on_message_listeners
      await self.__dispatch_helper(src=src,update=update,update_type=update_type)
    elif update.get("callback_query"):
      update_type, src = "callback_query", self.on_callback_query_listeners
      await self.__dispatch_helper(src=src,update=update,update_type=update_type)
    elif update.get("inline_query"):
      update_type, src = "inline_query", self.on_inline_query_listeners
      await self.__dispatch_helper(src=src,update=update,update_type=update_type)


from .dispatch_update import Dispatch
from .call import Call
from .create_message import CreateMessage
from .stop import Stop
from .start import Start
from .start_polling import StartPolling
from .webhook import Webhook

class Utilities(
  Dispatch,
  Call,
  CreateMessage,
  Stop,
  Start,
  StartPolling,
  Webhook
):
  pass


from aiohttp import web
import asyncio

class Webhook:
  async def takeCareWebhook(self, request):
    update = await request.json()
    if "update_id" in update:
      asyncio.create_task(self.dispatch_update(update))
    return web.json_response({"ok": True})
  
  async def createWebhook(self):
    if not self.connected:
      raise ConnectionError("Client is not connected. please start the client.")
    self.webhook, api, url = True, self.api, self.ApiUrl
    um = await api.post(url+"setWebhook", {"url": self.webhook_url+"/webhook"})
    if not um.get('ok'):
      raise Exception(f"Telegram says: {um}")
    app = web.Application()
    app.router.add_post("/webhook", self.takeCareWebhook)
    runner = web.AppRunner(app, access_log=None)
    self.log.info("Success. now listening updates from webhook.")
    await runner.setup()
    site = web.TCPSite(runner, "0.0.0.0", self.webhook_port)
    await site.start()


from .send_message import sendMessage
from .forward_messages import ForwardMessages
from .copy_messages import CopyMessages

class Messages(
  sendMessage,
  ForwardMessages,
  CopyMessages,
):
  pass


import asyncio
from Nexgram.errors import *

class StartPolling:
  async def start_polling(self):
    if not self.connected:
      raise ConnectionError("Client is not connected. Please connect the client and start polling.")
    elif self.polling: raise PollingAlreadyStartedError("Polling already started, why you trying again and again? didn't you receive any updates?")
    self.polling, log = True, self.log
    log.info("Nexgram.py - polling started!")
    first_start = True
    max_retry, retry = 25, 1
    while self.polling:
      try:
        params = {"offset": self.offset, "timeout": 15}
        updates = await self.api.get(self.ApiUrl+"getUpdates", params=params)
        if "result" in updates and not first_start:
          for update in updates["result"]:
            self.offset = update["update_id"] + 1
            asyncio.create_task(self.dispatch_update(update))
        elif "result" in updates and first_start:
          first_start = False
        elif 'error_code' in updates:
          err = updates.get('description')
          code = int(updates.get('error_code'))
          if code == 401:
            raise BadRequest(f"Telegram says: [{code}_{err.upper()}] - The bot token you provied is incorrect or revoked (caused by 'Client.start_polling')")
      except Exception as e:
        if retry > max_retry:
          break
        log.error(f"[{retry}/{max_retry}] Error in polling: {e}")
        retry += 1
    await self.stop()


from Nexgram.types import User
from Nexgram.errors import *

class GetChatMember:
  async def get_chat_member(self, chat_id: int, user_id: int):
    api, log, url = self.api, self.log, self.ApiUrl
    d = {"chat_id": chat_id, "user_id": user_id}
    r = await api.post(url+"getChatMember", d)
    if r.get('ok') and r.get('result'):
      user = r['result']['user']
      return r
      return User(
        client=self,
        id=user.get('id'),
        first_name=user.get('first_name'),
        last_name=user.get('last_name'),
        username=user.get('username'),
        is_bot=user.get('is_bot'),
      )
    return False
    if not r.get('ok') and r.get('error_code'):
      error_type = r.get('description')
      error = r.get('description').split(':', 1)[1]
      if 'bad request' in error_type.lower():
        raise BadRequest(error)
    # incompleted